/*
 * Segment 4 Datenübernahme und Aufbereitung (DUA), SWE 4.DELzFh DE Langzeit-Fehlererkennung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:<br>
 * BitCtrl Systems GmbH<br>
 * Weißenfelser Straße 67<br>
 * 04229 Leipzig<br>
 * Phone: +49 341-490670<br>
 * mailto: info@bitctrl.de
 */

package de.bsvrz.dua.langfehlerlve.modell.ausw;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dua.langfehlerlve.modell.FahrzeugArt;
import de.bsvrz.dua.langfehlerlve.modell.Rechenwerk;
import de.bsvrz.dua.langfehlerlve.modell.online.IDELzFhDatenListener;
import de.bsvrz.dua.langfehlerlve.modell.online.IDELzFhDatum;
import de.bsvrz.dua.langfehlerlve.modell.online.Intervall;
import de.bsvrz.dua.langfehlerlve.modell.online.PublikationsKanal;
import de.bsvrz.dua.langfehlerlve.parameter.IMsgDatenartParameter;
import de.bsvrz.sys.funclib.bitctrl.dua.DUAInitialisierungsException;
import de.bsvrz.sys.funclib.bitctrl.dua.DUAKonstanten;
import de.bsvrz.sys.funclib.bitctrl.dua.lve.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.dua.lve.MessStelle;

/**
 * Diese Klasse korrespondiert mit einem DAV-Objekt vom Typ
 * <code>typ.messStelle</code> und kapselt direkt oder indirekt saemtliche
 * Funktionalitäten, die innerhalb der SWE DE Langzeit-Fehlererkennung in Bezug
 * auf Objekte dieses Typs benoetigt werden. Ggf. (bei Intervallende) werden die
 * hier aggregierten DELzFh-Werte an die Messstellengruppe zur Analyse
 * weitergereicht
 *
 * @author BitCtrl Systems GmbH, Thierfelder
 */
public class DELzFhMessStelle extends AbstraktDELzFhObjekt implements IDELzFhDatenListener, ClientSenderInterface {

	/**
	 * diese Messstelle.
	 */
	private MessStelle messStelle = null;

	/**
	 * Datenbeschreibung zur Veroeffentlichung von MQ-Daten fuer die
	 * Intervallverkehrsstaerke.
	 */
	private DataDescription mqDb = null;

	/**
	 * Datenbeschreibung zur Veroeffentlichung von MS-Daten fuer die
	 * Intervallverkehrsstaerke.
	 */
	private DataDescription msDb = null;

	/**
	 * Aktuelle Daten aller mit der Messstelle assoziierten Messquerschnitte.
	 */
	private Map<SystemObject, Intervall> mqPuffer = new HashMap<>();

	/**
	 * Menge von Beobachtern der Online-Daten dieses Objektes.
	 */
	private Set<IDELzFhDatenListener> listenerMenge = new HashSet<>();

	/**
	 * wenn dieser Wert auf <code>!= null</code> steht, bedeutet das, dass das
	 * letzte eingetroffene Datum fuer dieses Objekt das erste des (neuen)
	 * Intervalls I + 1 ist und das mindestens ein Datum fuer das Intervall I
	 * vorhanden ist. In diesem Objekt stehen die Intervallgrenzen.
	 */
	private Intervall fertigesIntervall = null;

	/**
	 * Publikationskanal fuer MQ-Daten.
	 */
	private PublikationsKanal mqKanal = null;

	/**
	 * Publikationskanal fuer MS-Daten.
	 */
	private PublikationsKanal msKanal = null;

	/**
	 * Standardkonstruktor.
	 *
	 * @param dav               Datenverteiler-Verbindung
	 * @param ms                Systemobjekt vom Typ <code>typ.messStelle</code>
	 * @param messStellenGruppe die mit diesem Objekt assoziierte Messstellengruppe
	 * @param langZeit          Indiziert, ob sich dieses Objekt um das
	 *                          Langzeit-Vergleichsintervall kuemmern soll
	 * @throws Exception wenn das Objekt nicht sinnvoll initialisiert werden konnte
	 */
	protected DELzFhMessStelle(ClientDavInterface dav, MessStelle ms, DELzFhMessStellenGruppe messStellenGruppe,
			final boolean langZeit) throws Exception {
		messStelle = ms;

		mqKanal = new PublikationsKanal(dav);
		msKanal = new PublikationsKanal(dav);

		mqDb = new DataDescription(dav.getDataModel().getAttributeGroup("atg.intervallVerkehrsStärke"),
				dav.getDataModel().getAspect("asp.messQuerschnittKurzZeit"));
		msDb = new DataDescription(dav.getDataModel().getAttributeGroup("atg.intervallVerkehrsStärke"),
				dav.getDataModel().getAspect("asp.messStelleKurzZeit"));
		if (langZeit) {
			mqDb = new DataDescription(dav.getDataModel().getAttributeGroup("atg.intervallVerkehrsStärke"),
					dav.getDataModel().getAspect("asp.messQuerschnittLangZeit"));
			msDb = new DataDescription(dav.getDataModel().getAttributeGroup("atg.intervallVerkehrsStärke"),
					dav.getDataModel().getAspect("asp.messStelleLangZeit"));
		}

		super.init(dav, messStellenGruppe, langZeit);

		if (messStelle != null) {

			dav.subscribeSender(this, messStelle.getSystemObject(), mqDb, SenderRole.source());
			dav.subscribeSender(this, messStelle.getSystemObject(), msDb, SenderRole.source());

			if (messStelle.getPruefling() != null) {
				messStellenGruppe.getMq(messStelle.getPruefling().getSystemObject()).addListener(this);

				messStelle.getAbfahrten().stream()
				.forEach((abfahrt) -> messStellenGruppe.getMq(abfahrt.getSystemObject()).addListener(this));
				messStelle.getZufahrten().stream()
				.forEach((zufahrt) -> messStellenGruppe.getMq(zufahrt.getSystemObject()).addListener(this));
				initMQPuffer();
			} else {
				System.out.println(messStelle.getPruefling());
				throw new DUAInitialisierungsException("Messstelle " + ms + " besitzt keinen Pruefling (MQ)");
			}
		} else {
			throw new DUAInitialisierungsException("Messstelle konnte nicht ausgelesen werden");
		}
	}

	/**
	 * Erfragt das Strukturobjekt dieser Messstelle.
	 *
	 * @return das Strukturobjekt dieser Messstelle
	 */
	public final MessStelle getMessStelle() {
		return messStelle;
	}

	/**
	 * Initialisiert bzw. loescht den MQ-Puffer
	 */
	private void initMQPuffer() {
		synchronized (mqPuffer) {
			mqPuffer.put(messStelle.getPruefling().getSystemObject(), null);

			messStelle.getAbfahrten().stream()
			.forEach((abfahrt) -> mqPuffer.put(abfahrt.getSystemObject(), null));
			messStelle.getZufahrten().stream()
			.forEach((zufahrt) -> mqPuffer.put(zufahrt.getSystemObject(), null));
		}
	}

	/**
	 * Fuegt diesem Objekt einen neuen Listener hinzu und informiert diesen ggf.
	 * ueber aktuelle Daten
	 *
	 * @param listener eine neuer Listener
	 */
	public final void addListener(final IDELzFhDatenListener listener) {
		synchronized (listenerMenge) {
			if (listenerMenge.add(listener) && (fertigesIntervall != null)) {
				listener.aktualisiereDatum(messStelle.getSystemObject(), fertigesIntervall);
			}
		}
	}

	@Override
	public void dataRequest(SystemObject object, DataDescription dataDescription, byte state) {
		// Quellenanmeldung
	}

	@Override
	public boolean isRequestSupported(SystemObject object, DataDescription dataDescription) {
		return false;
	}

	/**
	 * Untersucht, ob eine Berechnung des Q-Wertes fuer Messstellen eingeleitet
	 * werden kann (durch das empfangene Datum getriggert). Dies ist der Fall, wenn
	 * ein Datum
	 *
	 * @param mqObjekt   ein Systemobjekt des MQ, fuer den Daten empfangen wurden
	 * @param neuesDatum ein MQ-Datum mit Nutzdaten
	 */
	private synchronized void versucheMessStellenBerechnung(SystemObject mqObjekt, Intervall neuesDatum) {
		synchronized (mqPuffer) {
			if (mqPuffer.size() > 0) {
				if (mqPuffer.size() > 1) {
					int schonEingetroffeneIntervalle = 0;

					for (Entry<SystemObject, Intervall> entry : mqPuffer.entrySet()) {
						Intervall aktuellesIntervall = entry.getValue();
						if (aktuellesIntervall != null) {
							if (neuesDatum.getStart() > aktuellesIntervall.getStart()) {
								veroeffentlicheAktuellenMsWert();
								mqPuffer.put(mqObjekt, neuesDatum);
								return;
							}
							if (neuesDatum.getStart() == aktuellesIntervall.getStart()) {
								schonEingetroffeneIntervalle++;
								// } else {
								// throw new RuntimeException(
								// "Veralteten Zeitstempel empfangen: " +
								// mqObjekt);
							}
						}
					}
					mqPuffer.put(mqObjekt, neuesDatum);

					if ((mqPuffer.size() - 1) == schonEingetroffeneIntervalle) {
						veroeffentlicheAktuellenMsWert();
					}
				} else {
					mqPuffer.put(mqObjekt, neuesDatum);
					veroeffentlicheAktuellenMsWert();
				}
			}
		}
	}

	/**
	 * Berechnet und veroeffentlicht ein MS-Datum auf Basis der im MQ-Puffer
	 * gespeicherten Daten. Danach werde alle gepufferten Daten wieder gelöscht
	 */
	private void veroeffentlicheAktuellenMsWert() {
		IDELzFhDatum pDatum = null;
		Set<IDELzFhDatum> abfahrtsDaten = new HashSet<>();
		Set<IDELzFhDatum> zufahrtsDaten = new HashSet<>();

		Intervall intervall = null;

		for (Entry<SystemObject, Intervall> entry : mqPuffer.entrySet()) {
			SystemObject mq = entry.getKey();
			Intervall aktuellesIntervall = entry.getValue();
			if (aktuellesIntervall != null) {
				intervall = aktuellesIntervall;

				boolean brk = false;
				if (messStelle.getPruefling().getSystemObject().equals(mq)) {
					pDatum = aktuellesIntervall.getDatum();
					brk = true;
				}

				if (!brk) {
					for (MessQuerschnittAllgemein zufahrt : messStelle.getZufahrten()) {
						if (zufahrt.getSystemObject().equals(mq)) {
							zufahrtsDaten.add(aktuellesIntervall.getDatum());
							brk = true;
							break;
						}
					}
				}

				if (!brk) {
					for (MessQuerschnittAllgemein abfahrt : messStelle.getAbfahrten()) {
						if (abfahrt.getSystemObject().equals(mq)) {
							abfahrtsDaten.add(aktuellesIntervall.getDatum());
							break;
						}
					}
				}
			}
		}

		IDELzFhDatum ergebnis = Rechenwerk.subtrahiere(Rechenwerk.addiere(pDatum, Rechenwerk.addiere(zufahrtsDaten)),
				Rechenwerk.addiere(abfahrtsDaten));

		if (intervall != null) {
			fertigesIntervall = new Intervall(intervall.getStart(), intervall.getEnde(), ergebnis);

			Data nutzDaten = AbstraktDELzFhObjekt.dDav.createData(msDb.getAttributeGroup());
			FahrzeugArt.getInstanzen().stream().forEach((fahrzeugArt) -> {
				if (fertigesIntervall.getDatum().getQ(fahrzeugArt) < 0.0) {
					nutzDaten.getUnscaledValue(fahrzeugArt.getAttributName())
					.set(DUAKonstanten.NICHT_ERMITTELBAR_BZW_FEHLERHAFT);
				} else {
					nutzDaten.getUnscaledValue(fahrzeugArt.getAttributName())
					.set(Math.round(fertigesIntervall.getDatum().getQ(fahrzeugArt)));
				}
			});

			ResultData msIntervallVSResultat = new ResultData(messStelle.getSystemObject(), msDb,
					fertigesIntervall.getStart(), nutzDaten);
			msKanal.publiziere(msIntervallVSResultat);

			synchronized (listenerMenge) {
				listenerMenge.stream().forEach((listener) -> listener
						.aktualisiereDatum(messStelle.getSystemObject(), fertigesIntervall));
			}
		}

		/**
		 * Datenpuffer initialisieren
		 */
		initMQPuffer();
	}

	@Override
	public void aktualisiereDatum(SystemObject mqObjekt, Intervall intervallDatum) {
		if (intervallDatum.getDatum().isKeineDaten()) {
			ResultData msIntervallVSResultat = new ResultData(messStelle.getSystemObject(), msDb,
					intervallDatum.getStart(), null);
			fertigesIntervall = intervallDatum;
			synchronized (listenerMenge) {
				listenerMenge.stream().forEach((listener) -> listener
						.aktualisiereDatum(messStelle.getSystemObject(), fertigesIntervall));
			}
			msKanal.publiziere(msIntervallVSResultat);
		} else {
			versucheMessStellenBerechnung(mqObjekt, intervallDatum);
		}

		if (messStelle.getPruefling().getSystemObject().equals(mqObjekt)) {
			Data nutzDaten = null;
			if (!intervallDatum.getDatum().isKeineDaten()) {
				nutzDaten = AbstraktDELzFhObjekt.dDav.createData(mqDb.getAttributeGroup());
				for (FahrzeugArt fahrzeugArt : FahrzeugArt.getInstanzen()) {
					if (intervallDatum.getDatum().getQ(fahrzeugArt) < 0.0) {
						nutzDaten.getUnscaledValue(fahrzeugArt.getAttributName())
						.set(DUAKonstanten.NICHT_ERMITTELBAR_BZW_FEHLERHAFT);
					} else {
						nutzDaten.getUnscaledValue(fahrzeugArt.getAttributName())
						.set(Math.round(intervallDatum.getDatum().getQ(fahrzeugArt)));
					}
				}
			}

			ResultData mqIntervallVSResultat = new ResultData(messStelle.getSystemObject(), mqDb,
					intervallDatum.getStart(), nutzDaten);
			mqKanal.publiziere(mqIntervallVSResultat);
		}
	}

	@Override
	protected void aktualisiereMsgParameter(IMsgDatenartParameter parameter) {
		initMQPuffer();
	}

}
